'use strict';

var url = require('url');


function HttpProbe(agent) {
  this.agent = agent;

  this.packages = ['http', 'https'];
}
exports.HttpProbe = HttpProbe;

function finalizeTransaction(err, profiler, time, transaction, res) {
    if(!time.done()) return;

    transaction.statusCode = res && res.statusCode || 500;
    transaction.stackTrace = profiler.formatStackTrace(err);
    transaction.error = err;

    if((!transaction.error) && ((transaction.statusCode < 200) || (transaction.statusCode >= 400))) {
      transaction.error = new Error("Error code, " + transaction.statusCode + ", no further details");
    }

    profiler.endTransaction(time, transaction);
}

function createBTCallback(proxy, profiler, time, transaction, res, thread, callback, self, args) {
    var didRun = false;
    var threadId = thread.current();

    return function () {
        if (didRun) return;
        didRun = true;

        var oldThreadId = thread.current();
        thread.resume(threadId);
        try {
            callback = proxy.wrapWithThreadProxyIfEnabled(callback);
            callback.apply(self, args);
        } catch (e) {
            finalizeTransaction(e, profiler, time, transaction);
            throw e;
        } finally {
            thread.resume(oldThreadId);
        }
    };
}

HttpProbe.prototype.__createRequestHandler = function (callback, isHTTPs) {
    var self = this;

    return function (req, res) {
        var profiler = self.agent.profiler;
        var proxy = self.agent.proxy;
        var time = profiler.time(true);

        var transaction = profiler.startTransaction(time, req, 'NODEJS_WEB');

        transaction.url = req.url;
        transaction.method = req.method;
        transaction.requestHeaders = req.headers;

        if(self.agent.eum.enabled && !transaction.corrHeader) {
          proxy.before(res, 'writeHead', function(obj, args) {
            var eumCookie = self.agent.eum.newEumCookie(transaction, req, obj, isHTTPs);
            eumCookie.build();
          });
        }

        proxy.after(res, 'end', function(obj, args) {
            finalizeTransaction(null, profiler, time, transaction, res);
        });

        if (self.agent.opts.btEntryPointDelayDisabled) {
            try {
                return callback.apply(this, arguments);
            } catch (e) {
                finalizeTransaction(e, profiler, time, transaction);
                throw e;
            }
        }

        var delayedCallback = createBTCallback(self.agent.proxy,
                                               profiler,
                                               time,
                                               transaction,
                                               res,
                                               self.agent.thread,
                                               callback,
                                               this,
                                               arguments);
        transaction.once('btInfoResponse', delayedCallback);
        setTimeout(delayedCallback, 10);
    };
}


HttpProbe.prototype.attach = function(obj, moduleName) {
  var self = this;

  if(obj.__appdynamicsProbeAttached__) return;
  obj.__appdynamicsProbeAttached__ = true;

  var profiler = self.agent.profiler;
  var proxy = self.agent.proxy;

  self.isHTTPs = obj.Agent && (obj.Agent.prototype.defaultPort == 443);

  // server probe
  self.agent.proxy.before(obj.Server.prototype, ['on', 'addListener'], function(obj, args) {
    if(args[0] !== 'request') return;

    if(obj.__httpProbe__) return;
    obj.__httpProbe__ = true;

    var cbIndex = args.length -1 ;

    args[cbIndex] = self.__createRequestHandler(args[cbIndex], moduleName === 'https');
  });



  // client probe
  function clientCallback(locals) {
    if(!locals.time.done()) return;

    var exitCall = locals.exitCall;
    var error = locals.error;

    if(exitCall) {
      if(locals.res) {
        exitCall.responseHeaders = locals.res.headers;
        exitCall.statusCode = ~~locals.res.statusCode;
        if((!error) && ((exitCall.statusCode < 200) || (exitCall.statusCode >= 400))) {
          error = new Error("Error code, " + exitCall.statusCode + ", no further details");
        }
      }

      profiler.addExitCall(locals.time, exitCall, error);
    }
  }

  // support 0.11.x and further
  if(obj.globalAgent && obj.globalAgent.request) {
    obj = obj.globalAgent;
  }

  proxy.around(obj, 'request', function(obj, args, locals) {
    if(typeof(args[0]) === 'string') {
      locals.opts = url.parse(args[0]);
    }
    else {
      locals.opts = args[0];
    }
    locals.time = profiler.time();

    var protocol = (moduleName === 'https' ? 'https:' : 'http:')
    var host = locals.opts.hostname || locals.opts.host;
    var port = locals.opts.port || (moduleName === 'https' ? 443 : 80);

    locals.exitCall = profiler.createExitCall(locals.time, {
      exitType: 'EXIT_HTTP',
      backendName: 'HTTP',
      label: protocol + '//' + host + ':' + port,
      identifyingProperties: {
        HOST: host,
        PORT: port
      },
      stackTrace: profiler.stackTrace(),
      group: (locals.opts.method || 'GET'),
      method: locals.opts.method,
      command: host +':' + port +(locals.opts.path || '/'),
      requestHeaders: locals.opts.headers,
      category: ((locals.opts.method === 'POST' || locals.opts.method === 'PUT') ? "write" : "read"),
    });

    proxy.callback(args, -1, function(obj, args) {
      var res = locals.res = args[0];
      proxy.before(res, ['on', 'addListener'], function(obj, args) {
        // workaround for end event
        if(args[0] === 'data' && obj.on !== undefined) {
          obj.on('end', function() {});
        }

        if(args[0] !== 'end') return;

        clientCallback(locals);
      });
    });
  },
  function(obj, args, ret, locals) {
    var writeOnce = false;

    proxy.before(ret, ['write', 'end'], function(obj, args) {
      if(!writeOnce) {
        writeOnce = true;
      }
      else {
        return;
      }

      if(locals.exitCall) {
        var transaction = profiler.getTransaction(locals.exitCall.threadId);
        if (transaction) {
          var corrHeader = self.agent.correlation.newCorrelationHeader();
          corrHeader.build(transaction, locals.exitCall);
          obj.setHeader(self.agent.correlation.HEADER_NAME, corrHeader.getStringHeader());

          // to check if we are ignoring possible snapshot request from btInfoResponse
          if(!transaction.btInfoResponse) {
            self.agent.logger.log("btInfoResponse is not yet available");
          }
        }
      }
    });

    proxy.before(ret, ['on', 'addListener'], function(obj, args) {
      if(args[0] == 'response') {
        proxy.callback(args, -1, function(obj, args) {
          locals.res = args[0];
          clientCallback(locals);
        });
      }
      else if(args[0] == 'error') {
        proxy.callback(args, -1, function(obj, args) {
          locals.error = args[0];
          clientCallback(locals);
        });
      }
    });
  });
};
